@@ -279,15 +279,20 @@ class PMApt(PackageManager):
 
 
 class PMDnf(PackageManager):
+    """
+    This is "legacy" DNF, called DNF-4 even though the
+    executable is dnf-3 in modern Fedora. Executable dnf
+    is a symlink to dnf-3 in systems that use it.
+    """
     backend = "dnf"
 
     def install(self, pkgs, from_local=False):
-        check_target_env_call(["dnf", "-y", "install"] + pkgs)
+        check_target_env_call(["dnf-3", "-y", "install"] + pkgs)
 
     def remove(self, pkgs):
         # ignore the error code for now because dnf thinks removing a
         # nonexistent package is an error
-        target_env_call(["dnf", "--disablerepo=*", "-C", "-y",
+        target_env_call(["dnf-3", "--disablerepo=*", "-C", "-y",
                          "remove"] + pkgs)
 
     def update_db(self):
@@ -295,7 +300,31 @@ class PMDnf(PackageManager):
         pass
 
     def update_system(self):
-        check_target_env_call(["dnf", "-y", "upgrade"])
+        check_target_env_call(["dnf-3", "-y", "upgrade"])
+
+
+class PMDnf5(PackageManager):
+    """
+    This is "modern" DNF, DNF-5 which is for Fedora 41 (presumably)
+    and later. Executable dnf is a symlink to dnf5 in systems that use it.
+    """
+    backend = "dnf5"
+
+    def install(self, pkgs, from_local=False):
+        check_target_env_call(["dnf5", "-y", "install"] + pkgs)
+
+    def remove(self, pkgs):
+        # ignore the error code for now because dnf thinks removing a
+        # nonexistent package is an error
+        target_env_call(["dnf5", "--disablerepo=*", "-C", "-y",
+                         "remove"] + pkgs)
+
+    def update_db(self):
+        # Doesn't need updates
+        pass
+
+    def update_system(self):
+        check_target_env_call(["dnf5", "-y", "upgrade"])
 
 
 class PMDummy(PackageManager):
@@ -337,6 +366,21 @@ class PMEntropy(PackageManager):
         # Doesn't need to update the system explicitly
         pass
 
+class PMFlatpak(PackageManager):
+    backend = "flatpak"
+
+    def install(self, pkgs, from_local=False):
+        check_target_env_call(["flatpak", "install", "--assumeyes"] + pkgs)
+
+    def remove(self, pkgs):
+        check_target_env_call(["flatpak", "uninstall", "--noninteractive"] + pkgs)
+
+    def update_db(self):
+        pass
+
+    def update_system(self):
+        # Doesn't need to update the system explicitly
+        pass
 
 class PMLuet(PackageManager):
     backend = "luet"
@@ -379,9 +423,10 @@ class PMPacman(PackageManager):
     def __init__(self):
         import re
         progress_match = re.compile("^\\((\\d+)/(\\d+)\\)")
+
         def line_cb(line):
             if line.startswith(":: "):
-                self.in_package_changes = "package changes" in line
+                self.in_package_changes = "package" in line or "hooks" in line
             else:
                 if self.in_package_changes and line.endswith("...\n"):
                     # Update the message, untranslated; do not change the
@@ -391,35 +436,217 @@ class PMPacman(PackageManager):
                     global custom_status_message
                     custom_status_message = "pacman: " + line.strip()
                     libcalamares.job.setprogress(self.progress_fraction)
-                    libcalamares.utils.debug(line)
+            libcalamares.utils.debug(line)
 
         self.in_package_changes = False
         self.line_cb = line_cb
 
+        pacman = libcalamares.job.configuration.get("pacman", None)
+        if pacman is None:
+            pacman = dict()
+        if type(pacman) is not dict:
+            libcalamares.utils.warning("Job configuration *pacman* will be ignored.")
+            pacman = dict()
+        self.pacman_num_retries = pacman.get("num_retries", 0)
+        self.pacman_disable_timeout = pacman.get("disable_download_timeout", False)
+        self.pacman_needed_only = pacman.get("needed_only", False)
+
     def reset_progress(self):
         self.in_package_changes = False
         # These are globals
         self.progress_fraction = (completed_packages * 1.0 / total_packages)
 
+    def run_pacman(self, command, callback=False):
+        """
+        Call pacman in a loop until it is successful or the number of retries is exceeded
+        :param command: The pacman command to run
+        :param callback: An optional boolean that indicates if this pacman run should use the callback
+        :return:
+        """
+
+        pacman_count = 0
+        while pacman_count <= self.pacman_num_retries:
+            pacman_count += 1
+            try:
+                if False: # callback:
+                    libcalamares.utils.target_env_process_output(command, self.line_cb)
+                else:
+                    libcalamares.utils.target_env_process_output(command)
+
+                return
+            except subprocess.CalledProcessError:
+                if pacman_count <= self.pacman_num_retries:
+                    pass
+                else:
+                    raise
+
     def install(self, pkgs, from_local=False):
+        command = ["pacman"]
+
         if from_local:
-            pacman_flags = "-U"
+            command.append("-U")
         else:
-            pacman_flags = "-S"
+            command.append("-S")
+
+        # Don't ask for user intervention, take the default action
+        command.append("--noconfirm")
+
+        # Don't report download progress for each file
+        command.append("--noprogressbar")
+
+        if self.pacman_needed_only is True:
+            command.append("--needed")
+
+        if self.pacman_disable_timeout is True:
+            command.append("--disable-download-timeout")
+
+        command += pkgs
 
         self.reset_progress()
-        libcalamares.utils.target_env_process_output(["pacman", pacman_flags,
-                               "--noconfirm"] + pkgs, self.line_cb)
+        self.run_pacman(command, True)
 
     def remove(self, pkgs):
         self.reset_progress()
-        libcalamares.utils.target_env_process_output(["pacman", "-Rs", "--noconfirm"] + pkgs, self.line_cb)
+        self.run_pacman(["pacman", "-Rs", "--noconfirm"] + pkgs, True)
 
     def update_db(self):
-        check_target_env_call(["pacman", "-Sy"])
+        self.run_pacman(["pacman", "-Sy"])
 
     def update_system(self):
-        check_target_env_call(["pacman", "-Su", "--noconfirm"])
+        command = ["pacman", "-Su", "--noconfirm"]
+        if self.pacman_disable_timeout is True:
+            command.append("--disable-download-timeout")
+
+        self.run_pacman(command)
+
+class PMParu(PackageManager):
+    backend = "paru"
+
+    def __init__(self):
+        import re
+        import shutil
+
+        progress_match = re.compile("^\\((\\d+)/(\\d+)\\)")
+
+        # Note that this writes to /etc/sudoers.d, NOT /etc/sudoers!
+        sudoers_entry = open("/etc/sudoers.d/autoupdate", "w")
+        sudoers_entry.write("nobody ALL = NOPASSWD: " + shutil.which("pacman"))
+        sudoers_entry.close()
+
+        def line_cb(line):
+            if line.startswith(":: "):
+                self.in_package_changes = "package" in line or "hooks" in line
+            else:
+                if self.in_package_changes and line.endswith("...\n"):
+                    # Update the message, untranslated; do not change the
+                    # progress percentage, since there may be more "installing..."
+                    # lines in the output for the group, than packages listed
+                    # explicitly. We don't know how to calculate proper progress.
+                    global custom_status_message
+                    custom_status_message = "paru: " + line.strip()
+                    libcalamares.job.setprogress(self.progress_fraction)
+            libcalamares.utils.debug(line)
+
+        self.in_package_changes = False
+        self.line_cb = line_cb
+
+        paru = libcalamares.job.configuration.get("paru", None)
+        if paru is None:
+            paru = dict()
+        if type(paru) is not dict:
+            libcalamares.utils.warning("Job configuration *paru* will be ignored.")
+            paru = dict()
+        self.paru_num_retries = paru.get("num_retries", 0)
+        self.paru_disable_timeout = paru.get("disable_download_timeout", False)
+        self.paru_needed_only = paru.get("needed_only", False)
+
+    def reset_progress(self):
+        self.in_package_changes = False
+        # These are globals
+        self.progress_fraction = (completed_packages * 1.0 / total_packages)
+
+    def run_paru(self, command, callback=False):
+        """
+        Call paru in a loop until it is successful or the number of retries is exceeded
+        :param command: The paru command to run
+        :param callback: An optional boolean that indicates if this paru run should use the callback
+        :return:
+        """
+
+        paru_count = 0
+        while paru_count <= self.paru_num_retries:
+            paru_count += 1
+            try:
+                if False: # callback:
+                    libcalamares.utils.target_env_process_output(command, self.line_cb)
+                else:
+                    libcalamares.utils.target_env_process_output(command)
+
+                return
+            except subprocess.CalledProcessError:
+                if paru_count <= self.paru_num_retries:
+                    pass
+                else:
+                    raise
+
+    def install(self, pkgs, from_local=False):
+        import os
+
+        os.environ["PWD"] = "/var/cache"
+        os.environ["XDG_CACHE_HOME"] = "/var/cache"
+
+        command = ["sudo", "-E", "-u", "nobody", "paru"]
+
+        if from_local:
+            command.append("-U")
+        else:
+            command.append("-S")
+
+        # Don't ask for user intervention, take the default action
+        command.append("--noconfirm")
+
+        # Don't report download progress for each file
+        command.append("--noprogressbar")
+
+        if self.paru_needed_only is True:
+            command.append("--needed")
+
+        if self.paru_disable_timeout is True:
+            command.append("--disable-download-timeout")
+
+        command += pkgs
+
+        self.reset_progress()
+        self.run_paru(command, True)
+
+    def remove(self, pkgs):
+        import os
+
+        os.environ["PWD"] = "/var/cache"
+        os.environ["XDG_CACHE_HOME"] = "/var/cache"
+
+        self.reset_progress()
+        self.run_paru(["sudo", "-E", "-u", "nobody", "paru", "-Rs", "--noconfirm"] + pkgs, True)
+
+    def update_db(self):
+        import os
+
+        os.environ["PWD"] = "/var/cache"
+        os.environ["XDG_CACHE_HOME"] = "/var/cache"
+
+        self.run_paru(["sudo", "-E", "-u", "nobody", "paru", "-Sy"])
+
+    def update_system(self):
+        import os
+
+        os.environ["PWD"] = "/var/cache"
+        os.environ["XDG_CACHE_HOME"] = "/var/cache"
+
+        command = ["sudo", "-E", "-u", "nobody", "paru", "-Su", "--noconfirm"]
+        if self.paru_disable_timeout is True:
+            command.append("--disable-download-timeout")
+
+        self.run_paru(command)
 
 
 class PMPamac(PackageManager):
@@ -485,17 +712,23 @@ class PMPortage(PackageManager):
 class PMXbps(PackageManager):
     backend = "xbps"
 
+    def line_cb(self, line):
+        libcalamares.utils.debug(line)
+
+    def run_xbps(self, command):
+        libcalamares.utils.target_env_process_output(command, self.line_cb);
+
     def install(self, pkgs, from_local=False):
-        check_target_env_call(["xbps-install", "-Sy"] + pkgs)
+        self.run_xbps(["xbps-install", "-Sy"] + pkgs)
 
     def remove(self, pkgs):
-        check_target_env_call(["xbps-remove", "-Ry", "--noconfirm"] + pkgs)
+        self.run_xbps(["xbps-remove", "-Ry"] + pkgs)
 
     def update_db(self):
-        check_target_env_call(["xbps-install", "-S"])
+        self.run_xbps(["xbps-install", "-S"])
 
     def update_system(self):
-        check_target_env_call(["xbps", "-Suy"])
+        self.run_xbps(["xbps", "-Suy"])
 
 
 class PMYum(PackageManager):
